"""This module implements the SqrtISwapGate."""
from __future__ import annotations

import math

from bqskit.ir.gates.constantgate import ConstantGate
from bqskit.ir.gates.qubitgate import QubitGate
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix


class SqrtISwapGate(ConstantGate, QubitGate):
    """
    The square root two qubit swap and phase iSWAP gate.

    The SqrtISwap gate is given by the following unitary:

    .. math::

        \\begin{pmatrix}
        1 & 0 & 0 & 0 \\\\
        0 & \\frac{1}{\\sqrt{2}} & \\frac{i}{\\sqrt{2}} & 0 \\\\
        0 & \\frac{i}{\\sqrt{2}} & \\frac{1}{\\sqrt{2}} & 0 \\\\
        0 & 0 & 0 & 1 \\\\
        \\end{pmatrix}
    """

    _num_qudits = 2
    _qasm_name = 'sqisw'
    _utry = UnitaryMatrix(
        [
            [1, 0, 0, 0],
            [0, 1 / math.sqrt(2), 1j / math.sqrt(2), 0],
            [0, 1j / math.sqrt(2), 1 / math.sqrt(2), 0],
            [0, 0, 0, 1],
        ],
    )



"""This module implements the MiddleOutLayerGenerator class."""
from __future__ import annotations

import itertools as it
import logging

from bqskit.compiler.basepass import BasePass
from bqskit.compiler.passdata import PassData
from bqskit.ir.circuit import Circuit
from bqskit.passes.search.generators.seed import SeedLayerGenerator
from bqskit.passes.search.generators.simple import SimpleLayerGenerator

_logger = logging.getLogger(__name__)


# TODO: choose a better name?
class MiddleOutLayerGenerator(SimpleLayerGenerator):
    """Layer Generator for search that adds gates at each time step."""
    generated_circuits: set[int]

    def __init__(self) -> None:
        """Construct a MiddleOutLayerGenerator."""
        self.generated_circuits = set()
        super().__init__()

    def gen_successors(self, circuit: Circuit, data: PassData) -> list[Circuit]:
        """
        Generate the successors of a circuit node.

        Raises:
            ValueError: If circuit is a single-qudit circuit.
        """

        if not isinstance(circuit, Circuit):
            raise TypeError('Expected circuit, got %s.' % type(circuit))

        if circuit.num_qudits < 2:
            raise ValueError('Cannot expand a single-qudit circuit.')

        # Get the machine model
        coupling_graph = BasePass.get_connectivity(circuit, data)

        # Generate successors
        successors = []
        for (edge, cycle) in it.product(
                coupling_graph, range(circuit.num_cycles),
        ):
            successor = circuit.copy()
            successor.insert_gate(
                cycle, self.two_qudit_gate, [
                    edge[0], edge[1],
                ],
            )
            successor.insert_gate(cycle, self.single_qudit_gate_1, edge[0])
            successor.insert_gate(cycle, self.single_qudit_gate_2, edge[1])
            # TODO: evaluate whether or not this hash method is efficient
            hash = SeedLayerGenerator.hash_structure(successor)
            if hash not in self.generated_circuits:
                self.generated_circuits.add(hash)
                successors.append(successor)

        for edge in coupling_graph:
            successor = circuit.copy()
            successor.append_gate(self.two_qudit_gate, [edge[0], edge[1]])
            successor.append_gate(self.single_qudit_gate_1, edge[0])
            successor.append_gate(self.single_qudit_gate_2, edge[1])
            if hash not in self.generated_circuits:
                self.generated_circuits.add(hash)
                successors.append(successor)

        return successors



from __future__ import annotations

from abc import ABC
from abc import abstractmethod

from bqskit.exec.results import RunnerResults
from bqskit.ir.circuit import Circuit


class CircuitRunner(ABC):
    """A CircuitRunner is responsible for executing a quantum circuit."""

    @abstractmethod
    def run(self, circuit: Circuit) -> RunnerResults:
        """
        Execute the circuit and return results.

        Args:
            circuit (Circuit): The circuit to run.

        Returns:
            (RunnerResults): The results from executing this circuit.
        """



"""This module implements the U1qGate."""
from __future__ import annotations

import math

import numpy as np
import numpy.typing as npt

from bqskit.ir.gates.qubitgate import QubitGate
from bqskit.qis.unitary.differentiable import DifferentiableUnitary
from bqskit.qis.unitary.unitary import RealVector
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix
from bqskit.utils.cachedclass import CachedClass


class U1qGate(QubitGate, DifferentiableUnitary, CachedClass):
    """
    The Quantinuum U1q single qubit gate.

    It is given by the following parameterized unitary:

    .. math::

        \\begin{pmatrix}
        \\cos{\\frac{\\theta_0}{2}} &
         -i\\exp{-i\\theta_1}\\sin{\\frac{\\theta_0}{2}} \\\\
        -i\\exp{i\\theta_1}\\sin{\\frac{\\theta_0}{2}} &
         \\cos{\\frac{\\theta_0}{2}} \\\\
        \\end{pmatrix}
    """

    _num_qudits = 1
    _num_params = 2
    _qasm_name = 'U1q'

    def get_unitary(self, params: RealVector = []) -> UnitaryMatrix:
        """Return the unitary for this gate, see :class:`Unitary` for more."""
        self.check_parameters(params)

        ct = np.cos(params[0] / 2)
        st = np.sin(params[0] / 2)
        enp = np.exp(-1j * params[1])
        epp = np.exp(1j * params[1])

        return UnitaryMatrix(
            [
                [ct, -1j * enp * st],
                [-1j * epp * st, ct],
            ],
        )

    def get_grad(self, params: RealVector = []) -> npt.NDArray[np.complex128]:
        """
        Return the gradient for this gate.

        See :class:`DifferentiableUnitary` for more info.
        """
        self.check_parameters(params)

        ct = np.cos(params[0] / 2)
        st = np.sin(params[0] / 2)
        enp = np.exp(-1j * params[1])
        epp = np.exp(1j * params[1])

        return np.array(
            [
                [  # wrt params[0]
                    [-0.5 * st, -0.5j * enp * ct],
                    [-0.5j * epp * ct, -0.5 * st],
                ],

                [  # wrt params[1]
                    [0, -1 * enp * st],
                    [epp * st, 0],
                ],
            ], dtype=np.complex128,
        )


U1qPiGate = U1qGate().with_frozen_params({0: math.pi})
U1qPi2Gate = U1qGate().with_frozen_params({0: math.pi / 2})



"""This module defines a standard `compile` function using BQSKit."""
from __future__ import annotations

import logging
import math
import warnings
from typing import Any
from typing import Literal
from typing import overload
from typing import Sequence
from typing import TYPE_CHECKING
from typing import Union

from bqskit.compiler.compiler import Compiler
from bqskit.compiler.machine import MachineModel
from bqskit.compiler.passdata import PassData
from bqskit.compiler.registry import _compile_circuit_registry
from bqskit.compiler.registry import _compile_statemap_registry
from bqskit.compiler.registry import _compile_stateprep_registry
from bqskit.compiler.registry import _compile_unitary_registry
from bqskit.compiler.registry import model_registered_target_types
from bqskit.compiler.workflow import Workflow
from bqskit.compiler.workflow import WorkflowLike
from bqskit.ir.circuit import Circuit
from bqskit.ir.gates.barrier import BarrierPlaceholder
from bqskit.ir.gates.constant.sx import SqrtXGate
from bqskit.ir.gates.measure import MeasurementPlaceholder
from bqskit.ir.gates.parameterized.rz import RZGate
from bqskit.ir.gates.parameterized.u3 import U3Gate
from bqskit.ir.operation import Operation
from bqskit.ir.opt import HilbertSchmidtCostGenerator
from bqskit.ir.opt import ScipyMinimizer
from bqskit.ir.opt.minimizers.lbfgs import LBFGSMinimizer
from bqskit.passes.control.foreach import ForEachBlockPass
from bqskit.passes.control.foreach import gen_replace_filter
from bqskit.passes.control.foreach import ReplaceFilterFn
from bqskit.passes.control.ifthenelse import IfThenElsePass
from bqskit.passes.control.predicates.change import ChangePredicate
from bqskit.passes.control.predicates.count import GateCountPredicate
from bqskit.passes.control.predicates.many import ManyQuditGatesPredicate
from bqskit.passes.control.predicates.multi import MultiPhysicalPredicate
from bqskit.passes.control.predicates.notpredicate import NotPredicate
from bqskit.passes.control.predicates.single import AllConstantSingleQuditGates
from bqskit.passes.control.predicates.single import HasGeneralSingleQuditGate
from bqskit.passes.control.predicates.single import NoSingleQuditGatesInModel
from bqskit.passes.control.predicates.single import SinglePhysicalPredicate
from bqskit.passes.control.predicates.single import ZXGatePredicate
from bqskit.passes.control.predicates.width import WidthPredicate
from bqskit.passes.control.whileloop import WhileLoopPass
from bqskit.passes.group import PassGroup
from bqskit.passes.mapping.apply import ApplyPlacement
from bqskit.passes.mapping.embed import EmbedAllPermutationsPass
from bqskit.passes.mapping.layout.pam import PAMLayoutPass
from bqskit.passes.mapping.layout.sabre import GeneralizedSabreLayoutPass
from bqskit.passes.mapping.placement.greedy import GreedyPlacementPass
from bqskit.passes.mapping.routing.pam import PAMRoutingPass
from bqskit.passes.mapping.routing.sabre import GeneralizedSabreRoutingPass
from bqskit.passes.mapping.setmodel import ExtractModelConnectivityPass
from bqskit.passes.mapping.setmodel import RestoreModelConnectivityPass
from bqskit.passes.mapping.setmodel import SetModelPass
from bqskit.passes.mapping.topology import SubtopologySelectionPass
from bqskit.passes.mapping.verify import PAMVerificationSequence
from bqskit.passes.measure import ExtractMeasurements
from bqskit.passes.measure import RestoreMeasurements
from bqskit.passes.noop import NOOPPass
from bqskit.passes.partitioning.quick import QuickPartitioner
from bqskit.passes.partitioning.single import GroupSingleQuditGatePass
from bqskit.passes.processing.scan import ScanningGateRemovalPass
from bqskit.passes.retarget.auto import AutoRebase2QuditGatePass
from bqskit.passes.retarget.general import GeneralSQDecomposition
from bqskit.passes.rules.u3 import U3Decomposition
from bqskit.passes.rules.zxzxz import ZXZXZDecomposition
from bqskit.passes.search.generators.single import SingleQuditLayerGenerator
from bqskit.passes.search.heuristics.dijkstra import DijkstraHeuristic
from bqskit.passes.synthesis.leap import LEAPSynthesisPass
from bqskit.passes.synthesis.pas import PermutationAwareSynthesisPass
from bqskit.passes.synthesis.qsearch import QSearchSynthesisPass
from bqskit.passes.synthesis.synthesis import SynthesisPass
from bqskit.passes.synthesis.target import SetTargetPass
from bqskit.passes.util.extend import ExtendBlockSizePass
from bqskit.passes.util.fill import FillSingleQuditGatesPass
from bqskit.passes.util.log import LogErrorPass
from bqskit.passes.util.log import LogPass
from bqskit.passes.util.random import SetRandomSeedPass
from bqskit.passes.util.unfold import UnfoldPass
from bqskit.qis.state import StateLike
from bqskit.qis.state.state import StateVector
from bqskit.qis.state.system import StateSystem
from bqskit.qis.state.system import StateSystemLike
from bqskit.qis.unitary import UnitaryLike
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix
from bqskit.utils.typing import is_integer
from bqskit.utils.typing import is_iterable
from bqskit.utils.typing import is_real_number

if TYPE_CHECKING:
    from bqskit.compiler.basepass import BasePass


_logger = logging.getLogger(__name__)


CompilationInputLike = Union[Circuit, UnitaryLike, StateLike, StateSystemLike]
CompilationInput = Union[Circuit, UnitaryMatrix, StateVector, StateSystem]


@overload
def compile(
    input: CompilationInputLike,
    model: MachineModel | None = ...,
    *,
    with_mapping: Literal[False] = ...,
    optimization_level: int = ...,
    max_synthesis_size: int = ...,
    synthesis_epsilon: float = ...,
    error_threshold: float | None = ...,
    error_sim_size: int = ...,
    compiler: Compiler | None = ...,
    seed: int | None = ...,
    **compiler_kwargs: Any,
) -> Circuit:
    ...


@overload
def compile(
    input: Sequence[CompilationInput],
    model: MachineModel | None = ...,
    *,
    with_mapping: Literal[False] = ...,
    optimization_level: int = ...,
    max_synthesis_size: int = ...,
    synthesis_epsilon: float = ...,
    error_threshold: float | None = ...,
    error_sim_size: int = ...,
    compiler: Compiler | None = ...,
    seed: int | None = ...,
    **compiler_kwargs: Any,
) -> list[Circuit]:
    ...


@overload
def compile(
    input: CompilationInputLike,
    model: MachineModel | None = ...,
    *,
    with_mapping: Literal[True],
    optimization_level: int = ...,
    max_synthesis_size: int = ...,
    synthesis_epsilon: float = ...,
    error_threshold: float | None = ...,
    error_sim_size: int = ...,
    compiler: Compiler | None = ...,
    seed: int | None = ...,
    **compiler_kwargs: Any,
) -> tuple[Circuit, tuple[int, ...], tuple[int, ...]]:
    ...


@overload
def compile(
    input: Sequence[CompilationInput],
    model: MachineModel | None = ...,
    *,
    with_mapping: Literal[True],
    optimization_level: int = ...,
    max_synthesis_size: int = ...,
    synthesis_epsilon: float = ...,
    error_threshold: float | None = ...,
    error_sim_size: int = ...,
    compiler: Compiler | None = ...,
    seed: int | None = ...,
    **compiler_kwargs: Any,
) -> list[tuple[Circuit, tuple[int, ...], tuple[int, ...]]]:
    ...


@overload
def compile(
    input: CompilationInputLike,
    model: MachineModel | None = ...,
    *,
    with_mapping: bool,
    optimization_level: int = ...,
    max_synthesis_size: int = ...,
    synthesis_epsilon: float = ...,
    error_threshold: float | None = ...,
    error_sim_size: int = ...,
    compiler: Compiler | None = ...,
    seed: int | None = ...,
    **compiler_kwargs: Any,
) -> Circuit | tuple[Circuit, tuple[int, ...], tuple[int, ...]]:
    ...


@overload
def compile(
    input: Sequence[CompilationInput],
    model: MachineModel | None = ...,
    *,
    with_mapping: bool,
    optimization_level: int = ...,
    max_synthesis_size: int = ...,
    synthesis_epsilon: float = ...,
    error_threshold: float | None = ...,
    error_sim_size: int = ...,
    compiler: Compiler | None = ...,
    seed: int | None = ...,
    **compiler_kwargs: Any,
) -> list[Circuit] | list[tuple[Circuit, tuple[int, ...], tuple[int, ...]]]:
    ...


def compile(
    input: CompilationInputLike | Sequence[CompilationInput],
    model: MachineModel | None = None,
    optimization_level: int = 1,
    max_synthesis_size: int = 3,
    synthesis_epsilon: float = 1e-8,
    error_threshold: float | None = None,
    error_sim_size: int = 8,
    compiler: Compiler | None = None,
    seed: int | None = None,
    with_mapping: bool = False,
    **compiler_kwargs: Any,
) -> (
    Circuit
    | tuple[Circuit, tuple[int, ...], tuple[int, ...]]
    | list[Circuit]
    | list[tuple[Circuit, tuple[int, ...], tuple[int, ...]]]
):
    """
    Compile a circuit, unitary, or state with a standard workflow.

    Args:
        input (CompilationInputLike | Sequence[CompilationInput]): The input
            or inputs to compile. If a single input is given, a single
            circuit will be returned. If an iterable of inputs is given,
            a list of circuits will be returned.

        model (MachineModel | None): A model of the target machine.
            Defaults to an all-to-all connected hardware with CNOTs and U3s
            as native gates. See :class:`MachineModel` for information
            on loading a preset model or creating a custom one.

        optimization_level (int): The degree of optimization in the workflow.
            The workflow will produce better circuits at the cost of
            performance with a higher number.  An optimization_level
            of 0 is not supported due to inherit optimization in any
            workflow containing synthesis. Valid inputs are 1, 2, 3, or 4.
            (Default: 1)

        max_synthesis_size (int): The maximum size of a unitary to
            synthesize or instantiate. Larger circuits will be partitioned.
            Increasing this will most likely lead to better results with
            an exponential time trade-off. (Default: 3)

        synthesis_epsilon (float): The maximum distance between target
            and circuit unitary during any instantiation or synthesis
            algorithms. (Default: 1e-8)

        error_threshold (float | None): This parameter controls the
            verification mechanism in this compile function. By default,
            it is set to None, so no verification is done. If you set
            this to a float, the upper bound on compilation error
            is calculated. If the upper bound is larger than this number,
            a warning will be logged. (Default: None)

        error_sim_size (int): If an `error_threshold` is set, the error
            upper bound is calculated by simulating blocks of this size.
            As you increase `error_sim_size`, the upper bound on error
            becomes more accurate. This setting is ignored with direct
            synthesis compilations, i.e., when a state, system, or unitary
            are given as input. (Default: 8)

        compiler (Compiler | None): Pass a :class:`Compiler` to prevent
            creating one. Save on startup time by passing a compiler in
            when calling `compile` multiple times. (Default: None)

        seed (int | None): Set a seed for the compile function for
            better reproducibility. If left as None, will not set seed.

        with_mapping (bool): If True, three values will be returned
            instead of just the compiled circuit. The first value is the
            compiled circuit, the second value is the initial mapping,
            and the third value is the final mapping. The initial mapping
            is a tuple where `initial_mapping[i] = j` implies that logical
            qudit `i` in the input system starts on the physical qudit
            `j` in the output circuit. Likewise, the final mapping describes
            where the logical qudits are in the physical circuit at the end
            of execution. (Default: False)

        compiler_kwargs (Any): Passed directly to BQSKit compiler construction.
            Arguments for connecting to a cluster can go here.
            See :class:`Compiler` for more info.

    Returns:
        (Circuit): The compiled circuit.

    Examples:
        >>> from bqskit import Circuit, compile
        >>> circuit = Circuit.from_file('input.qasm')
        >>> compiled_circuit = compile(circuit)
        >>> compiled_circuit.save('output.qasm')

        >>> from bqskit import compile
        >>> from bqskit.qis.state import StateVector
        >>> state = StateVector.random(3)
        >>> compiled_circuit = compile(state)
        >>> compiled_circuit.save('output.qasm')

        >>> from bqskit import compile
        >>> from bqskit.qis.unitary import UnitaryMatrix
        >>> utry = UnitaryMatrix.random(3)
        >>> compiled_circuit = compile(utry)
        >>> compiled_circuit.save('output.qasm')

        >>> from bqskit import compile, Circuit
        >>> circuits = [Circuit.from_file(f'input{i}.qasm') for i in range(5)]
        >>> compiled_circuits = compile(circuits)
        >>> for i, circuit in enumerate(compiled_circuits):
        ...     circuit.save(f'output{i}.qasm')

        >>> from bqskit import compile, Circuit, MachineModel
        >>> from bqskit.ir.gates import CZGate, RZGate, SqrtXGate
        >>> target_gate_set = {CZGate(), RZGate(), SqrtXGate()}
        >>> circuit = Circuit.from_file('input.qasm')
        >>> model = MachineModel(circuit.num_qudits, gate_set=target_gate_set)
        >>> compiled_circuit = compile(circuit, model, optimization_level=2)

        You can also use pre-built models from the :obj:`~bqskit.ext` package
        for common hardware. For example, to compile to the H1-1 machine
        from Quantinuum:

        >>> from bqskit import compile, Circuit
        >>> from bqskit.ext import H1_1Model
        >>> circuit = Circuit.from_file('input.qasm')
        >>> compiled_circuit = compile(circuit, H1_1Model)

    Raises:
        ValueError: If the input is an empty iterable.

        ValueError: If the input (or any input) is larger than the model.

        ValueError: If either the input or the model has mixed radixes.

        ValueError: If the model has a mismatched radix with the circuit.

        ValueError: If the model doesn't contain any entangling gates and
            the input (or any input) is larger than one-qudit.

        ValueError: If the model does not contain any entangling gates
            that are less than or equal to the input size.

        ValueError: If optimization_level is not 1, 2, 3, or 4.

        ValueError: If the maximum synthesis size is less than the
            largest gate in the model.

        ValueError: If the maximum synthesis size is less than 2.

        ValueError: If the synthesis epsilon is not between 0 and 1.

        ValueError: The error simulation size is less than the maximum
            synthesis size.
    """
    # Check `model`
    if model is not None and not isinstance(model, MachineModel):
        raise TypeError(f'Expected MachineModel for model, got {type(model)}.')

    # Check `optimization_level`
    if not is_integer(optimization_level):
        raise TypeError(
            'Expected integer for optimization_level'
            f', got{type(optimization_level)}.',
        )

    if optimization_level < 1 or optimization_level > 4:
        raise ValueError(
            'The optimization level should be either 1, 2, 3, or 4.'
            f' Got {optimization_level}.',
        )

    # Check `max_synthesis_size`
    if not is_integer(max_synthesis_size):
        raise TypeError(
            'Expected integer for max_synthesis_size'
            f', got{type(max_synthesis_size)}.',
        )

    if max_synthesis_size <= 1:
        raise ValueError(
            'The maximum synthesis size must be greater than or equal to 2.'
            f' Got {max_synthesis_size}.',
        )

    if model is not None:
        if max_synthesis_size < max(g.num_qudits for g in model.gate_set):
            value = max(g.num_qudits for g in model.gate_set)
            raise ValueError(
                'The maximum synthesis size needs to be greater than or equal'
                ' to the largest native gate size: '
                f'{max_synthesis_size} < {value}.',
            )

    if max_synthesis_size > 6:
        _logger.warning('Large max synthesis size; expect long compile times.')

    # Check `synthesis_epsilon`
    if not is_real_number(synthesis_epsilon):
        raise TypeError(
            'Expected float for synthesis_epsilon'
            f', got {type(synthesis_epsilon)}.',
        )

    if synthesis_epsilon < 0 or synthesis_epsilon > 1:
        raise ValueError(
            'Out-of-bounds synthesis_epsilon, it must be between 0 and 1,'
            f' got {synthesis_epsilon}.',
        )

    # Check `error_threshold`
    if error_threshold is not None:
        if not is_real_number(error_threshold):
            raise TypeError(
                'Expected float for error_threshold'
                f', got {type(error_threshold)}.',
            )

        if error_threshold < 0 or error_threshold > 1:
            raise ValueError(
                'Out-of-bounds error_threshold, it must be between 0 and 1,'
                f' got {error_threshold}.',
            )

    # Check `error_sim_size`
    if not is_integer(error_sim_size):
        raise TypeError(
            'Expected integer for error_sim_size'
            f', got{type(error_sim_size)}.',
        )

    if error_sim_size < max_synthesis_size:
        raise ValueError(
            'Simulation size for error calculation cannot be less than the'
            f'maximum synthesis size: {error_sim_size} < {max_synthesis_size}.',
        )

    # Check `compiler`
    if compiler is not None and not isinstance(compiler, Compiler):
        raise TypeError(
            'Expected Compiler object for compiler parameter'
            f', got {type(compiler)}.',
        )

    # Check `seed`
    if seed is not None and not is_integer(seed):
        raise TypeError(f'Expected integer for seed, got {type(seed)}.')

    # check `with_mapping`
    if not isinstance(with_mapping, bool):
        raise TypeError(
            f'Expected bool for with_mapping, got {type(with_mapping)}.',
        )

    def type_and_check_input(input: CompilationInputLike) -> CompilationInput:
        """Check input to be valid and convert to a proper typed object."""
        try:
            if isinstance(input, Circuit):
                pass

            elif UnitaryMatrix.is_unitary(input):
                input = UnitaryMatrix(input)

            elif StateVector.is_pure_state(input):
                input = StateVector(input)

            elif StateSystem.is_state_system(input):
                input = StateSystem(input)

            else:
                raise TypeError(
                    'Input is neither a circuit, a unitary, a state system'
                    f', nor a state. Got {type(input)}.',
                )
        except Exception as e:
            raise TypeError(
                'Unable to determine type of input.'
                ' Ensure that you are trying to compile a valid'
                ' circuit, unitary, state, or state-system.',
            ) from e

        assert isinstance(
            input, (
                Circuit,
                UnitaryMatrix,
                StateVector,
                StateSystem,
            ),
        )

        if not all(r == input.radixes[0] for r in input.radixes):
            raise ValueError(
                'Currently, can only automatically build a workflow '
                'for same-level systems, such as qubit-only or qutrit-only'
                'systems. mixed-radix systems are not yet supported'
                'with the standard workflows.',
            )

        if model is not None:
            if model.num_qudits < input.num_qudits:
                raise ValueError('Machine is too small for this input.')

            if not all(r == input.radixes[0] for r in model.radixes):
                raise ValueError('Model has a mismatch in radixes with input.')

            if len(model.gate_set.multi_qudit_gates) == 0:
                if input.num_qudits > 1:
                    raise ValueError('No entangling gates in native gate set.')

            if all(g.num_qudits > input.num_qudits for g in model.gate_set):
                raise ValueError(
                    'Model gate set does not contain any entangling gates'
                    ' that is less than or equal to input size.'
                    f' Cannot compile {input.num_qudits}-qudit input without'
                    f' {input.num_qudits}-qudit or smaller gates.',
                )

        return input

    # Connect to or construct a Compiler
    managed_compiler = compiler is None

    if managed_compiler:
        compiler = Compiler(**compiler_kwargs)

    assert compiler is not None

    if (
        is_iterable(input)
        and not isinstance(input, Circuit)
        and not UnitaryMatrix.is_unitary(input)
        and not StateVector.is_pure_state(input)
        and not StateSystem.is_state_system(input)
    ):
        typed_inputs = [type_and_check_input(i) for i in input]  # type: ignore

        if len(typed_inputs) == 0:
            raise ValueError('Input iterable is empty.')

        # Build workflows
        workflows = [
            build_workflow(
                typed_input,
                model,
                optimization_level,
                synthesis_epsilon,
                max_synthesis_size,
                error_threshold,
                error_sim_size,
                seed,
            )
            for typed_input in typed_inputs
        ]
        in_circuits = [
            typed_input if isinstance(typed_input, Circuit) else Circuit(1)
            for typed_input in typed_inputs
        ]

        # Perform the compilations in parallel
        job_ids = [
            compiler.submit(in_circuit, workflow, True)
            for in_circuit, workflow in zip(in_circuits, workflows)
        ]
        results = [compiler.result(job_id) for job_id in job_ids]

        outs: list[Circuit] = []
        datas: list[PassData] = []
        for result in results:
            out, data = result
            outs.append(out)  # type: ignore
            datas.append(data)  # type: ignore

        # Log error if necessary
        if error_threshold is not None:
            for i, data in enumerate(datas):
                error = data.error
                nonsq_error = 1 - math.sqrt(max(1 - (error * error), 0))
                if nonsq_error > error_threshold:
                    warnings.warn(
                        'Upper bound on error is greater than set threshold:'
                        f' {nonsq_error} > {error_threshold}'
                        f' in compilation index {i}.',
                    )

        # Gather mapping data if necessary
        if with_mapping:
            pis = []
            pfs = []
            for typed_input, data in zip(typed_inputs, datas):
                default = list(range(typed_input.num_qudits))
                pi = data.get('initial_mapping', default)
                pf = data.get('final_mapping', default)
                pis.append(pi)
                pfs.append(pf)
            to_return: Any = list(zip(outs, pis, pfs))

        else:
            to_return = outs

    else:
        typed_input = type_and_check_input(input)  # type: ignore

        # Build workflow
        workflow = build_workflow(
            typed_input,
            model,
            optimization_level,
            synthesis_epsilon,
            max_synthesis_size,
            error_threshold,
            error_sim_size,
            seed,
        )
        if isinstance(typed_input, Circuit):
            in_circuit = typed_input

        elif isinstance(typed_input, UnitaryMatrix):
            in_circuit = Circuit.from_unitary(typed_input)

        else:
            in_circuit = Circuit(typed_input.num_qudits, typed_input.radixes)

        # Perform the compilation
        out, data = compiler.compile(in_circuit, workflow, True)

        # Log error if necessary
        if error_threshold is not None:
            error = data.error
            nonsq_error = 1 - math.sqrt(max(1 - (error * error), 0))
            if nonsq_error > error_threshold:
                warnings.warn(
                    'Upper bound on error is greater than set threshold:'
                    f' {nonsq_error} > {error_threshold}.',
                )

        # Gather mapping data if necessary
        if with_mapping:
            default = list(range(typed_input.num_qudits))
            pi = data.get('initial_mapping', default)
            pf = data.get('final_mapping', default)
            to_return = (out, pi, pf)

        else:
            to_return = out

    # Close managed compiler
    if managed_compiler:
        compiler.close()

    return to_return


def build_workflow(
    input: Circuit | UnitaryMatrix | StateVector | StateSystem,
    model: MachineModel | None = None,
    optimization_level: int = 1,
    synthesis_epsilon: float = 1e-8,
    max_synthesis_size: int = 3,
    error_threshold: float | None = None,
    error_sim_size: int = 8,
    seed: int | None = None,
) -> Workflow:
    """Build a BQSKit Off-the-Shelf workflow, see :func:`compile` for info."""
    if model is None:
        model = MachineModel(input.num_qudits, radixes=input.radixes)

    model_registered_types = model_registered_target_types(model)

    if isinstance(input, Circuit):
        if input.num_qudits > max_synthesis_size:
            if any(
                g.num_qudits > max_synthesis_size
                and not isinstance(
                    g, (
                        MeasurementPlaceholder,
                        BarrierPlaceholder,
                    ),
                )
                for g in input.gate_set
            ):
                raise ValueError(
                    'Unable to compile circuit with gate larger than'
                    ' max_synthesis_size.\nConsider adjusting it.',
                )
        # Use a registered workflow if model is found in the circuit registry
        # for a given optimization_level
        if model in _compile_circuit_registry:
            if optimization_level in _compile_circuit_registry[model]:
                return _compile_circuit_registry[model][optimization_level]
        elif len(model_registered_types) > 0:
            m = f'MachineModel {model} is registered for inputs of type in '
            m += f'{model_registered_types}, but input is {type(input)}. '
            m += f'You may need to register a Workflow for type {type(input)}.'
            warnings.warn(m)

        return _circuit_workflow(
            model,
            optimization_level,
            synthesis_epsilon,
            max_synthesis_size,
            error_threshold,
            error_sim_size,
            seed,
        )

    elif isinstance(input, UnitaryMatrix):
        if input.num_qudits > max_synthesis_size:
            raise ValueError(
                'Unable to compile unitary with size larger than'
                ' max_synthesis_size.\nConsider adjusting it.',
            )
        # Use a registered workflow if model is found in the unitary registry
        # for a given optimization_level
        if model in _compile_unitary_registry:
            if optimization_level in _compile_unitary_registry[model]:
                return _compile_unitary_registry[model][optimization_level]
        elif len(model_registered_types) > 0:
            m = f'MachineModel {model} is registered for inputs of type in '
            m += f'{model_registered_types}, but input is {type(input)}. '
            m += f'You may need to register a Workflow for type {type(input)}.'
            warnings.warn(m)

        return _synthesis_workflow(
            input,
            model,
            optimization_level,
            synthesis_epsilon,
            max_synthesis_size,
            error_threshold,
            error_sim_size,
            seed,
        )

    elif isinstance(input, StateVector):
        if input.num_qudits > max_synthesis_size:
            raise ValueError(
                'Unable to compile states with size larger than'
                ' max_synthesis_size.\nConsider adjusting it.',
            )
        # Use a registered workflow if model is found in the stateprep registry
        # for a given optimization_level
        if model in _compile_stateprep_registry:
            if optimization_level in _compile_stateprep_registry[model]:
                return _compile_stateprep_registry[model][optimization_level]
        elif len(model_registered_types) > 0:
            m = f'MachineModel {model} is registered for inputs of type in '
            m += f'{model_registered_types}, but input is {type(input)}. '
            m += f'You may need to register a Workflow for type {type(input)}.'
            warnings.warn(m)

        return _stateprep_workflow(
            input,
            model,
            optimization_level,
            synthesis_epsilon,
            max_synthesis_size,
            error_threshold,
            error_sim_size,
            seed,
        )

    elif isinstance(input, StateSystem):
        if input.num_qudits > max_synthesis_size:
            raise ValueError(
                'Unable to compile state systems with size larger than'
                ' max_synthesis_size.\nConsider adjusting it.',
            )
        # Use a registered workflow if model is found in the statemap registry
        # for a given optimization_level
        if model in _compile_statemap_registry:
            if optimization_level in _compile_statemap_registry[model]:
                return _compile_statemap_registry[model][optimization_level]
        elif len(model_registered_types) > 0:
            m = f'MachineModel {model} is registered for inputs of type in '
            m += f'{model_registered_types}, but input is {type(input)}. '
            m += f'You may need to register a Workflow for type {type(input)}.'
            warnings.warn(m)

        return _statemap_workflow(
            input,
            model,
            optimization_level,
            synthesis_epsilon,
            max_synthesis_size,
            error_threshold,
            error_sim_size,
            seed,
        )

    raise TypeError(f'Unexpected input type: {type(input)}.')


def _circuit_workflow(
    model: MachineModel,
    optimization_level: int = 1,
    synthesis_epsilon: float = 1e-8,
    max_synthesis_size: int = 3,
    error_threshold: float | None = None,
    error_sim_size: int = 8,
    seed: int | None = None,
) -> Workflow:
    """Build standard workflow for circuit compilation."""
    workflow_builders = [
        _opt1_workflow,
        _opt2_workflow,
        _opt3_workflow,
        _opt4_workflow,
    ]
    workflow_builder = workflow_builders[optimization_level - 1]
    workflow: list[BasePass] = [] if seed is None else [SetRandomSeedPass(seed)]
    workflow += [UnfoldPass(), ExtractMeasurements()]
    workflow += workflow_builder(
        model,
        synthesis_epsilon,
        max_synthesis_size,
        error_threshold,
        error_sim_size,
    )
    workflow += [RestoreMeasurements()]
    return Workflow(workflow, name='Off-the-Shelf Circuit Compilation')


def get_instantiate_options(optimization_level: int) -> dict[str, Any]:
    """
    Return good default instantiate options based on an optimization level.

    Args:
        optimization_level (int): The optimization level.

    Returns:
        (dict[str, Any]): The options.

    Raises:
        ValueError: If the optimization level is not 1, 2, 3, or 4.
    """
    if optimization_level == 1:
        return {
            'multistarts': 1,
            'ftol': 1e-6,
            'gtol': 1e-10,
            'diff_tol_r': 1e-4,
            'max_iters': 1000,
            'min_iters': 0,
        }

    elif optimization_level == 2:
        return {
            'multistarts': 2,
            'ftol': 5e-12,
            'gtol': 1e-14,
            'diff_tol_r': 1e-5,
            'max_iters': 10000,
            'min_iters': 100,
        }

    elif optimization_level == 3:
        return {
            'multistarts': 4,
            'ftol': 5e-16,
            'gtol': 1e-15,
            'diff_tol_r': 5e-5,
            'max_iters': 50000,
            'min_iters': 200,
        }

    elif optimization_level == 4:
        return {
            'multistarts': 8,
            'ftol': 5e-16,
            'gtol': 1e-15,
            'diff_tol_r': 1e-6,
            'max_iters': 100000,
            'min_iters': 1000,
        }

    else:
        raise ValueError(
            'Invalid optimization level, must be 1, 2, 3, or 4.'
            f' Got {optimization_level}.',
        )


def build_partitioning_workflow(
    inner_workflow: WorkflowLike,
    block_size: int = 3,
    error_sim_size: int | None = None,
    replace_filter_method: str = 'less-than-respecting-multi',
) -> Workflow:
    """
    Build standard partitioning workflow for circuit compilation.

    This includes additional steps than normally necessary for error upper
    bound analysis.

    Args:
        inner_workflow (WorkflowLike): The workflow to partition.

        block_size (int): The size of the blocks to partition into.

        error_sim_size (int | None): The size of the blocks to simulate
            errors on. If None, then no error analysis is performed.

        replace_filter_method (str): The method to use for replacing
            gates in the circuit. See :class:`ForEachBlockPass` for more.

    Returns:
        Workflow: The partitioning workflow.

    Raises:
        ValueError: If block_size < 2.

        ValueError: If error_sim_size < block_size.
    """
    if not is_integer(block_size):
        raise TypeError(
            f'Expected block_size to be int, got {type(block_size)}.',
        )

    if block_size < 2:
        raise ValueError(f'Expected block_size > 1, got {block_size}.')

    if error_sim_size is not None and not is_integer(block_size):
        raise TypeError(
            f'Expected int for error_sim_size, got {type(error_sim_size)}.',
        )

    if error_sim_size is not None and error_sim_size < block_size:
        raise ValueError(
            f'Expected error_sim_size >= block_size, got {error_sim_size}.',
        )

    pass_list = [QuickPartitioner(block_size), ExtendBlockSizePass()]

    if error_sim_size is None:
        pass_list += [
            ForEachBlockPass(
                inner_workflow,
                replace_filter=replace_filter_method,
            ),
        ]

    elif error_sim_size == block_size:
        pass_list += [
            ForEachBlockPass(
                inner_workflow,
                replace_filter=replace_filter_method,
                calculate_error_bound=True,
            ),
        ]

    else:
        pass_list += [
            QuickPartitioner(error_sim_size),
            ForEachBlockPass(
                ForEachBlockPass(
                    inner_workflow,
                    replace_filter=replace_filter_method,
                ),
                calculate_error_bound=True,
            ),
        ]

    pass_list += [UnfoldPass()]
    return Workflow(pass_list, name='Partitioning')


def build_standard_search_synthesis_workflow(
    optimization_level: int = 1,
    synthesis_epsilon: float = 1e-8,
) -> BasePass:
    """
    Build standard search-based synthesis pass for block-level compilation.

    Args:
        optimization_level (int): The optimization level. See :func:`compile`
            for more information.

        synthesis_epsilon (float): The maximum distance between target
            and circuit unitary allowed to declare successful synthesis.
            Set to 0 for exact synthesis. (Default: 1e-8)

    Returns:
        BasePass: The synthesis pass.

    Raises:
        ValueError: If the optimization level is not 1, 2, 3, or 4.

    Note:
        For larger circuits, this pass may take a very long time to run.
        If unitary synthesis is your ultimate goal, rather than circuit
        compilation, consider designing a custom instantiation-based
        synthesis method or using alternative synthesis techniques
        -- such as QFAST or QPredict -- for large unitaries.
    """
    if optimization_level not in [1, 2, 3, 4]:
        raise ValueError(
            'Invalid optimization level, must be 1, 2, 3, or 4.'
            f' Got {optimization_level}.',
        )

    if not is_real_number(synthesis_epsilon):
        raise TypeError(
            'Expected float for synthesis_epsilon'
            f', got {type(synthesis_epsilon)}.',
        )

    qsearch = QSearchSynthesisPass(
        success_threshold=synthesis_epsilon,
        instantiate_options=get_instantiate_options(optimization_level),
    )
    leap = LEAPSynthesisPass(
        success_threshold=synthesis_epsilon,
        min_prefix_size=[3, 4, 7, 9][optimization_level - 1],
        instantiate_options=get_instantiate_options(optimization_level),
    )
    return IfThenElsePass(WidthPredicate(3), qsearch, leap)


def build_multi_qudit_retarget_workflow(
    optimization_level: int = 1,
    synthesis_epsilon: float = 1e-8,
    max_synthesis_size: int = 3,
    error_threshold: float | None = None,
    error_sim_size: int = 8,
) -> Workflow:
    """
    Build standard workflow for circuit multi-qudit gate set retargeting.

    Notes:
        - This workflow assumes that SetModelPass will be run earlier in the
          full workflow and doesn't add it in here.

        - For the most part, circuit connectivity isn't a concern during
          retargeting. However, if the circuit contains many-qudit (>= 3)
          gates, then the workflow will not preserve connectivity during
          the decomposition of those gates. If your input contains many-qudit
          gates, consider following this with a mapping workflow.
    """

    core_retarget_workflow = [
        FillSingleQuditGatesPass(),
        IfThenElsePass(
            NotPredicate(MultiPhysicalPredicate()),
            IfThenElsePass(
                ManyQuditGatesPredicate(),
                [
                    ExtractModelConnectivityPass(),
                    build_standard_search_synthesis_workflow(
                        optimization_level,
                        synthesis_epsilon,
                    ),
                    RestoreModelConnectivityPass(),
                ],
                AutoRebase2QuditGatePass(3, 5, synthesis_epsilon),
            ),
            ScanningGateRemovalPass(
                success_threshold=synthesis_epsilon,
                collection_filter=_mq_gate_collection_filter,
                instantiate_options=get_instantiate_options(optimization_level),
            ),
        ),
    ]

    return Workflow(
        [
            IfThenElsePass(
                NotPredicate(WidthPredicate(2)),
                [
                    LogPass('Retargeting multi-qudit gates.'),
                    build_partitioning_workflow(
                        core_retarget_workflow,
                        max_synthesis_size,
                        None if error_threshold is None else error_sim_size,
                    ),
                ],
            ),
        ], name='Multi Qudit Retargeting',
    )


def build_single_qudit_retarget_workflow(
    optimization_level: int = 1,
    synthesis_epsilon: float = 1e-8,
    max_synthesis_size: int = 3,
    error_threshold: float | None = None,
    error_sim_size: int = 8,
) -> BasePass:
    """Build a pass to convert single-qudit-gates to the native gate set."""
    sq_synthesis = QSearchSynthesisPass(
        layer_generator=SingleQuditLayerGenerator(None, allow_repeats=True),
        heuristic_function=DijkstraHeuristic(),
        instantiate_options={
            'method': 'minimization',
            'minimizer': ScipyMinimizer(),
            'cost_fn_gen': HilbertSchmidtCostGenerator(),
        },
    )

    sq_gate_deletion = build_partitioning_workflow(
        ScanningGateRemovalPass(
            success_threshold=synthesis_epsilon,
            collection_filter=_sq_gate_collection_filter,
            instantiate_options=get_instantiate_options(optimization_level),
        ),
        max_synthesis_size,
        None if error_threshold is None else error_sim_size,
    )

    return Workflow(
        [
            IfThenElsePass(
                NotPredicate(SinglePhysicalPredicate()),
                [
                    LogPass('Retargeting single-qudit gates.'),
                    UnfoldPass(),
                    GroupSingleQuditGatePass(),
                    IfThenElsePass(
                        AllConstantSingleQuditGates(),
                        # TODO: Implement better retargeting techniques
                        # for constant gate sets
                        LogPass(
                            'The standard workflow with BQSKit may have trouble'
                            ' targeting gate sets containing no parameterized'
                            ' single-qudit gates.',
                            logging.WARNING,
                        ),
                    ),
                    IfThenElsePass(
                        NoSingleQuditGatesInModel(),
                        [
                            LogPass('Attempting to remove single-qudit gates.'),
                            sq_gate_deletion,
                            IfThenElsePass(
                                NotPredicate(SinglePhysicalPredicate()),
                                LogPass(
                                    'Unable to remove all single-qudit gates;'
                                    ' gate set may not be universal without'
                                    ' single-qudit gates. Consider changing'
                                    ' gate set, increasing optimization_level'
                                    ' or max_synthesis_size.',
                                    logging.WARNING,
                                ),
                            ),
                        ],
                        [
                            ForEachBlockPass(
                                IfThenElsePass(
                                    NotPredicate(SinglePhysicalPredicate()),
                                    [
                                        IfThenElsePass(
                                            HasGeneralSingleQuditGate(),
                                            GeneralSQDecomposition(),
                                            IfThenElsePass(
                                                ZXGatePredicate(),
                                                ZXZXZDecomposition(),
                                                sq_synthesis,
                                            ),
                                        ),
                                    ],
                                ),
                            ),
                        ],
                    ),
                    UnfoldPass(),
                ],
            ),
        ], name='Single Qudit Retargeting',
    )


def build_sabre_mapping_workflow() -> Workflow:
    """
    Build standard workflow for circuit mapping.

    Note:
        - This workflow assumes that SetModelPass will be run earlier in the
          full workflow and doesn't add it in here.

        - It also assumes that ApplyPlacement will be run later in the full
          workflow and doesn't add it in here.
    """
    return Workflow(
        [
            LogPass('Mapping circuit.'),
            GreedyPlacementPass(),
            GeneralizedSabreLayoutPass(),
            GeneralizedSabreRoutingPass(),
        ], name='SABRE Mapping',
    )


def build_seqpam_mapping_optimization_workflow(
    optimization_level: int = 4,
    synthesis_epsilon: float = 1e-8,
    num_layout_passes: int = 3,
    block_size: int = 3,
    error_sim_size: int | None = None,
) -> Workflow:
    """
    Build a Sequential-Permutation-Aware Mapping and Optimizing Workflow.

    Note:
        - This workflow assumes that SetModelPass will be run earlier in the
          full workflow and doesn't add it in here.

        - This will apply the placement found during the workflow. The
        resulting circuit will be physically mapped.

    Args:
        optimization_level (int): The optimization level. See :func:`compile`
            for more information.

        synthesis_epsilon (float): The maximum distance between target
            and circuit unitary allowed to declare successful synthesis.
            Set to 0 for exact synthesis. (Default: 1e-8)

        num_layout_passes (int): The number of layout forward and backward
            passes to run. See :class:`PamLayoutPass` for more information.
            (Default: 3)

        block_size (int): The size of the blocks to partition into.
            Warning, the number of permutation evaluated increases
            factorially and the difficulty of each permutation increases
            exponentially with this. (Default: 3)

        error_sim_size (int | None): The size of the blocks to simulate
            errors on. If None, then no error analysis is performed.
            (Default: None)

    Raises:
        ValueError: If block_size < 2.

        ValueError: If error_sim_size < block_size.
    """
    if not is_integer(block_size):
        raise TypeError(
            f'Expected block_size to be int, got {type(block_size)}.',
        )

    if block_size < 2:
        raise ValueError(f'Expected block_size > 1, got {block_size}.')

    if error_sim_size is not None and not is_integer(block_size):
        raise TypeError(
            f'Expected int for error_sim_size, got {type(error_sim_size)}.',
        )

    if error_sim_size is not None and error_sim_size < block_size:
        raise ValueError(
            f'Expected error_sim_size >= block_size, got {error_sim_size}.',
        )

    qsearch = QSearchSynthesisPass(
        success_threshold=synthesis_epsilon,
        instantiate_options=get_instantiate_options(optimization_level),
    )

    leap = LEAPSynthesisPass(
        success_threshold=synthesis_epsilon,
        min_prefix_size=9,
        instantiate_options=get_instantiate_options(optimization_level),
    )

    if error_sim_size is not None:
        post_pam_seq: BasePass = PAMVerificationSequence(error_sim_size)
    else:
        post_pam_seq = NOOPPass()

    return Workflow(
        IfThenElsePass(
            NotPredicate(WidthPredicate(2)),
            [
                LogPass('Caching permutation-aware synthesis results.'),
                ExtractModelConnectivityPass(),
                QuickPartitioner(block_size),
                ForEachBlockPass(
                    IfThenElsePass(
                        WidthPredicate(4),
                        EmbedAllPermutationsPass(
                            inner_synthesis=qsearch,
                            input_perm=True,
                            output_perm=False,
                            vary_topology=False,
                        ),
                        EmbedAllPermutationsPass(
                            inner_synthesis=leap,
                            input_perm=True,
                            output_perm=False,
                            vary_topology=False,
                        ),
                    ),
                ),
                LogPass('Preoptimizing with permutation-aware mapping.'),
                PAMRoutingPass(),
                post_pam_seq,
                UnfoldPass(),
                RestoreModelConnectivityPass(),

                LogPass('Recaching permutation-aware synthesis results.'),
                SubtopologySelectionPass(block_size),
                QuickPartitioner(block_size),
                ForEachBlockPass(
                    IfThenElsePass(
                        WidthPredicate(4),
                        EmbedAllPermutationsPass(
                            inner_synthesis=qsearch,
                            input_perm=False,
                            output_perm=True,
                            vary_topology=True,
                        ),
                        EmbedAllPermutationsPass(
                            inner_synthesis=leap,
                            input_perm=False,
                            output_perm=True,
                            vary_topology=True,
                        ),
                    ),
                ),
                LogPass('Performing permutation-aware mapping.'),
                ApplyPlacement(),
                PAMLayoutPass(num_layout_passes),
                PAMRoutingPass(0.1),
                post_pam_seq,
                ApplyPlacement(),
                UnfoldPass(),
            ],
        ),
        name='SeqPAM Mapping',
    )


def build_gate_deletion_optimization_workflow(
    optimization_level: int = 1,
    synthesis_epsilon: float = 1e-8,
    max_synthesis_size: int = 3,
    error_threshold: float | None = None,
    error_sim_size: int = 8,
    iterative: bool = False,
) -> Workflow:
    """Build standard workflow for circuit gate deletion optimization."""
    core_workflow = Workflow(
        [
            LogPass('Attempting to delete gates.'),
            build_partitioning_workflow(
                ScanningGateRemovalPass(
                    success_threshold=synthesis_epsilon,
                    instantiate_options=get_instantiate_options(
                        optimization_level,
                    ),
                ),
                max_synthesis_size,
                None if error_threshold is None else error_sim_size,
            ),
        ],
        name='Gate Deletion Optimization',
    )

    if iterative:
        return Workflow(
            WhileLoopPass(ChangePredicate(), core_workflow),
            name='Iterative Gate Deletion Optimization',
        )

    return core_workflow


def build_resynthesis_optimization_workflow(
    optimization_level: int = 1,
    synthesis_epsilon: float = 1e-8,
    max_synthesis_size: int = 3,
    error_threshold: float | None = None,
    error_sim_size: int = 8,
    iterative: bool = False,
) -> Workflow:
    """Build standard workflow for circuit resynthesis optimization."""
    core_workflow = Workflow(
        [
            LogPass('Resynthesizing blocks.'),
            build_partitioning_workflow(
                IfThenElsePass(
                    NotPredicate(WidthPredicate(2)),
                    build_standard_search_synthesis_workflow(
                        optimization_level,
                        synthesis_epsilon,
                    ),
                ),
                max_synthesis_size,
                None if error_threshold is None else error_sim_size,
            ),
        ], name='Resynthesis Optimization',
    )

    if iterative:
        return Workflow(
            [
                WhileLoopPass(
                    GateCountPredicate('multi'),
                    core_workflow,
                ),
            ], name='Iterative Resynthesis Optimization',
        )

    return core_workflow


def _opt1_workflow(
    model: MachineModel,
    synthesis_epsilon: float = 1e-8,
    max_synthesis_size: int = 3,
    error_threshold: float | None = None,
    error_sim_size: int = 8,
) -> list[BasePass]:
    """Build Optimization Level 1 workflow for circuit compilation."""
    return [
        # Initializing
        SetModelPass(model),

        build_multi_qudit_retarget_workflow(
            1,
            synthesis_epsilon,
            max_synthesis_size,
            error_threshold,
            error_sim_size,
        ),

        build_sabre_mapping_workflow(),

        build_multi_qudit_retarget_workflow(
            1,
            synthesis_epsilon,
            max_synthesis_size,
            error_threshold,
            error_sim_size,
        ),

        build_single_qudit_retarget_workflow(
            1,
            synthesis_epsilon,
            max_synthesis_size,
            error_threshold,
            error_sim_size,
        ),

        # Finalizing
        LogErrorPass(),
        ApplyPlacement(),
    ]


def _opt2_workflow(
    model: MachineModel,
    synthesis_epsilon: float = 1e-8,
    max_synthesis_size: int = 3,
    error_threshold: float | None = None,
    error_sim_size: int = 8,
) -> list[BasePass]:
    """Build Optimization Level 2 workflow for circuit compilation."""
    return [
        # Initializing
        SetModelPass(model),

        build_multi_qudit_retarget_workflow(
            2,
            synthesis_epsilon,
            max_synthesis_size,
            error_threshold,
            error_sim_size,
        ),

        build_sabre_mapping_workflow(),


        build_multi_qudit_retarget_workflow(
            2,
            synthesis_epsilon,
            max_synthesis_size,
            error_threshold,
            error_sim_size,
        ),

        build_single_qudit_retarget_workflow(
            2,
            synthesis_epsilon,
            max_synthesis_size,
            error_threshold,
            error_sim_size,
        ),

        build_gate_deletion_optimization_workflow(
            2,
            synthesis_epsilon,
            max_synthesis_size,
            error_threshold,
            error_sim_size,
        ),

        # Finalizing
        LogErrorPass(),
        ApplyPlacement(),
    ]


def _opt3_workflow(
    model: MachineModel,
    synthesis_epsilon: float = 1e-8,
    max_synthesis_size: int = 3,
    error_threshold: float | None = None,
    error_sim_size: int = 8,
) -> list[BasePass]:
    """Build Optimization Level 3 workflow for circuit compilation."""
    return [
        # Initializing
        SetModelPass(model),

        build_multi_qudit_retarget_workflow(
            3,
            synthesis_epsilon,
            max_synthesis_size,
            error_threshold,
            error_sim_size,
        ),

        build_sabre_mapping_workflow(),


        build_multi_qudit_retarget_workflow(
            3,
            synthesis_epsilon,
            max_synthesis_size,
            error_threshold,
            error_sim_size,
        ),

        build_resynthesis_optimization_workflow(
            3,
            synthesis_epsilon,
            max_synthesis_size,
            error_threshold,
            error_sim_size,
            True,
        ),

        build_single_qudit_retarget_workflow(
            3,
            synthesis_epsilon,
            max_synthesis_size,
            error_threshold,
            error_sim_size,
        ),

        build_gate_deletion_optimization_workflow(
            3,
            synthesis_epsilon,
            max_synthesis_size,
            error_threshold,
            error_sim_size,
            True,
        ),

        # Finalizing
        LogErrorPass(),
        ApplyPlacement(),
    ]


def _opt4_workflow(
    model: MachineModel,
    synthesis_epsilon: float = 1e-8,
    max_synthesis_size: int = 3,
    error_threshold: float | None = None,
    error_sim_size: int = 8,
) -> list[BasePass]:
    """Build optimization Level 4 workflow for circuit compilation."""
    if max_synthesis_size > 3:
        _logger.warning(
            'It is currently recommended to set max_synthesis_size to 3'
            ' for optimization level 4. This may change in the future.',
        )

    return [
        SetModelPass(model),

        build_seqpam_mapping_optimization_workflow(
            4,
            synthesis_epsilon,
            block_size=max_synthesis_size,
            error_sim_size=None if error_threshold is None else error_sim_size,
        ),

        build_multi_qudit_retarget_workflow(
            4,
            synthesis_epsilon,
            max_synthesis_size,
            error_threshold,
            error_sim_size,
        ),

        build_resynthesis_optimization_workflow(
            4,
            synthesis_epsilon,
            max_synthesis_size,
            error_threshold,
            error_sim_size,
            True,
        ),

        build_single_qudit_retarget_workflow(
            4,
            synthesis_epsilon,
            max_synthesis_size,
            error_threshold,
            error_sim_size,
        ),

        build_gate_deletion_optimization_workflow(
            4,
            synthesis_epsilon,
            max_synthesis_size,
            error_threshold,
            error_sim_size,
            True,
        ),

        # Finalizing
        LogErrorPass(),
    ]


def _synthesis_workflow(
    input: UnitaryMatrix,
    model: MachineModel,
    optimization_level: int = 1,
    synthesis_epsilon: float = 1e-8,
    max_synthesis_size: int = 3,
    error_threshold: float | None = None,
    error_sim_size: int = 8,
    seed: int | None = None,
) -> Workflow:
    """Build a workflow for unitary synthesis."""
    if error_threshold is not None:
        if error_threshold < synthesis_epsilon:
            raise ValueError(
                'When performing direct synthesis, the error threshold'
                ' cannot be less than the synthesis epsilon.',
            )

    sq_synthesis = QSearchSynthesisPass(
        layer_generator=SingleQuditLayerGenerator(None, allow_repeats=True),
        heuristic_function=DijkstraHeuristic(),
        instantiate_options={
            'method': 'minimization',
            'minimizer': ScipyMinimizer(),
            'cost_fn_gen': HilbertSchmidtCostGenerator(),
        } if input.radixes == (2,) else {},
    )

    qsearch = QSearchSynthesisPass(
        success_threshold=synthesis_epsilon,
        instantiate_options=get_instantiate_options(optimization_level),
    )

    leap = LEAPSynthesisPass(
        success_threshold=synthesis_epsilon,
        min_prefix_size=[3, 4, 7, 9][optimization_level - 1],
        instantiate_options=get_instantiate_options(optimization_level),
    )

    if optimization_level < 4:
        synthesis: BasePass = IfThenElsePass(WidthPredicate(3), qsearch, leap)

    else:
        in_synthesis = qsearch if input.num_qudits <= 3 else leap
        synthesis = PermutationAwareSynthesisPass(inner_synthesis=in_synthesis)

    if input.num_qudits == 1:
        synthesis = sq_synthesis

    scan = ScanningGateRemovalPass(
        success_threshold=synthesis_epsilon,
        instantiate_options=get_instantiate_options(optimization_level),
    )

    workflow: list[BasePass] = [] if seed is None else [SetRandomSeedPass(seed)]
    workflow += [
        SetModelPass(model),
        SetTargetPass(input),
        synthesis,
        build_single_qudit_retarget_workflow(
            optimization_level,
            synthesis_epsilon,
            max_synthesis_size,
            error_threshold,
            error_sim_size,
        ),
        scan if optimization_level >= 2 else NOOPPass(),
    ]

    return Workflow(workflow, name='Off-the-Shelf Unitary Synthesis')


def _stateprep_workflow(
    state: StateVector,
    model: MachineModel,
    optimization_level: int = 1,
    synthesis_epsilon: float = 1e-8,
    max_synthesis_size: int = 3,
    error_threshold: float | None = None,
    error_sim_size: int = 8,
    seed: int | None = None,
) -> Workflow:
    """Build a workflow for state preparation."""
    layer_gen = model.gate_set.build_mq_layer_generator()

    if optimization_level == 1:
        inst_ops = {
            'multistarts': 1,
            'method': 'minimization',
            'minimizer': LBFGSMinimizer(),
        }
        synthesis: SynthesisPass = LEAPSynthesisPass(
            success_threshold=synthesis_epsilon,
            layer_generator=layer_gen,
            instantiate_options=inst_ops,
            min_prefix_size=3,
            cost=HilbertSchmidtCostGenerator(),
        )

    elif optimization_level == 2:
        inst_ops = {
            'multistarts': 4,
            'method': 'minimization',
            'ftol': 5e-12,
            'gtol': 1e-14,
        }
        synthesis = LEAPSynthesisPass(
            success_threshold=synthesis_epsilon,
            layer_generator=layer_gen,
            instantiate_options=inst_ops,
            min_prefix_size=5,
        )

    elif optimization_level == 3:
        inst_ops = {
            'multistarts': 8,
            'method': 'minimization',
            'ftol': 5e-16,
            'gtol': 1e-15,
        }
        if state.num_qudits > 3:
            synthesis = LEAPSynthesisPass(
                success_threshold=synthesis_epsilon,
                layer_generator=layer_gen,
                instantiate_options=inst_ops,
                min_prefix_size=7,
            )
        else:
            synthesis = QSearchSynthesisPass(
                success_threshold=synthesis_epsilon,
                layer_generator=layer_gen,
                instantiate_options=inst_ops,
            )

    elif optimization_level == 4:
        inst_ops = {
            'multistarts': 8,
            'method': 'minimization',
            'ftol': 5e-16,
            'gtol': 1e-15,
        }
        if state.num_qudits > 3:
            in_synthesis = LEAPSynthesisPass(
                success_threshold=synthesis_epsilon,
                layer_generator=layer_gen,
                instantiate_options=inst_ops,
                min_prefix_size=7,
            )
        else:
            in_synthesis = QSearchSynthesisPass(  # type: ignore
                success_threshold=synthesis_epsilon,
                layer_generator=layer_gen,
                instantiate_options=inst_ops,
            )
        synthesis = PermutationAwareSynthesisPass(inner_synthesis=in_synthesis)

    scan = ScanningGateRemovalPass(
        success_threshold=synthesis_epsilon,
        instantiate_options=inst_ops,
        cost=HilbertSchmidtCostGenerator(),
    )

    workflow: list[BasePass] = [] if seed is None else [SetRandomSeedPass(seed)]
    workflow += [
        SetModelPass(model),
        SetTargetPass(state),
        synthesis,
        scan if optimization_level >= 2 else NOOPPass(),
    ]

    return Workflow(workflow, name='Off-the-Shelf State Synthesis')


def _statemap_workflow(
    state: StateSystem,
    model: MachineModel,
    optimization_level: int = 1,
    synthesis_epsilon: float = 1e-8,
    max_synthesis_size: int = 3,
    error_threshold: float | None = None,
    error_sim_size: int = 8,
    seed: int | None = None,
) -> Workflow:
    """Build a workflow for state preparation."""
    layer_gen = model.gate_set.build_mq_layer_generator()

    if optimization_level == 1:
        inst_ops = {
            'multistarts': 1,
            'method': 'minimization',
        }
        synthesis: SynthesisPass = LEAPSynthesisPass(
            success_threshold=synthesis_epsilon,
            layer_generator=layer_gen,
            instantiate_options=inst_ops,
            min_prefix_size=3,
        )

    elif optimization_level == 2:
        inst_ops = {
            'multistarts': 4,
            'method': 'minimization',
            'ftol': 5e-12,
            'gtol': 1e-14,
        }
        synthesis = LEAPSynthesisPass(
            success_threshold=synthesis_epsilon,
            layer_generator=layer_gen,
            instantiate_options=inst_ops,
            min_prefix_size=5,
        )

    elif optimization_level == 3:
        inst_ops = {
            'multistarts': 8,
            'method': 'minimization',
            'ftol': 5e-16,
            'gtol': 1e-15,
        }
        if state.num_qudits > 3:
            synthesis = LEAPSynthesisPass(
                success_threshold=synthesis_epsilon,
                layer_generator=layer_gen,
                instantiate_options=inst_ops,
                min_prefix_size=7,
            )
        else:
            synthesis = QSearchSynthesisPass(
                success_threshold=synthesis_epsilon,
                layer_generator=layer_gen,
                instantiate_options=inst_ops,
            )

    elif optimization_level == 4:
        inst_ops = {
            'multistarts': 8,
            'method': 'minimization',
            'ftol': 5e-16,
            'gtol': 1e-15,
        }
        if state.num_qudits > 3:
            in_synthesis = LEAPSynthesisPass(
                success_threshold=synthesis_epsilon,
                layer_generator=layer_gen,
                instantiate_options=inst_ops,
                min_prefix_size=7,
            )
        else:
            in_synthesis = QSearchSynthesisPass(  # type: ignore
                success_threshold=synthesis_epsilon,
                layer_generator=layer_gen,
                instantiate_options=inst_ops,
            )
        synthesis = PermutationAwareSynthesisPass(inner_synthesis=in_synthesis)

    scan = ScanningGateRemovalPass(
        success_threshold=synthesis_epsilon,
        instantiate_options=inst_ops,
    )

    workflow: list[BasePass] = [] if seed is None else [SetRandomSeedPass(seed)]
    workflow += [
        SetModelPass(model),
        SetTargetPass(state),
        synthesis,
        scan if optimization_level >= 2 else NOOPPass(),
    ]

    return Workflow(workflow, name='Off-the-Shelf State System Synthesis')


def _get_single_qudit_gate_rebase_pass(model: MachineModel) -> BasePass:
    """Build a pass to convert single-qudit-gates to the native gate set."""
    warnings.warn(
        'This single qudit rebase workflow has moved'
        ' to build_single_qudit_retarget_workflow. See Workflow for more'
        ' information. This function is deprecated and will be removed'
        ' in a future release.',
        DeprecationWarning,
    )

    sq_gates = [g for g in model.gate_set if g.num_qudits == 1]

    if len(sq_gates) == 0:
        return NOOPPass()

    if all(g.is_constant() for g in sq_gates):
        _logger.warn(
            'The standard workflow with BQSKit may have trouble'
            ' targeting gate sets containing no parameterized'
            ' single-qudit gates.',
        )
        # TODO: Implement better retargeting techniques for constant gate sets

    instantiate_options = {
        'method': 'minimization',
        'minimizer': ScipyMinimizer(),
        'cost_fn_gen': HilbertSchmidtCostGenerator(),
    }
    layer_generator = SingleQuditLayerGenerator(sq_gates, True)
    core_sq_rebase: BasePass = NOOPPass()
    if len(sq_gates) == 1:
        if sq_gates[0] == U3Gate():
            core_sq_rebase = U3Decomposition()

    elif len(sq_gates) >= 2:
        if RZGate() in sq_gates and SqrtXGate() in sq_gates:
            core_sq_rebase = ZXZXZDecomposition()

    if isinstance(core_sq_rebase, NOOPPass):
        core_sq_rebase = QSearchSynthesisPass(
            layer_generator=layer_generator,
            heuristic_function=DijkstraHeuristic(),
            instantiate_options=instantiate_options,
        )

    return PassGroup([
        IfThenElsePass(
            NotPredicate(SinglePhysicalPredicate()),
            [
                GroupSingleQuditGatePass(),
                ForEachBlockPass([
                    IfThenElsePass(
                        NotPredicate(SinglePhysicalPredicate()),
                        core_sq_rebase,
                    ),
                ]),
                UnfoldPass(),
            ],
        ),
    ])


def _gen_replace_filter(model: MachineModel) -> ReplaceFilterFn:
    """Generate a replace filter for use during the standard workflow."""
    warnings.warn(
        'The replace filter generated by _gen_replace_filter has moved'
        ' to gen_replace_filter in foreach. See ForEachBlockPass for more'
        ' information. This function is deprecated and will be removed'
        ' in a future release.',
        DeprecationWarning,
    )
    return gen_replace_filter('less-then-respecting-multi', model)


def _mq_gate_collection_filter(op: Operation) -> bool:
    """Return true if `op` is a multi-qudit operation."""
    return op.num_qudits > 1


def _sq_gate_collection_filter(op: Operation) -> bool:
    """Return true if `op` is a multi-qudit operation."""
    return op.num_qudits == 1



"""
=======================================================
A Runtime for the Compile-time (:mod:`bqskit.runtime`)
=======================================================

.. currentmodule:: bqskit.runtime

This package provides an execution environment for compiler passes,
enabling them to parallelize and distribute their workload efficiently.

Launching a Server
##################

To parallelize BQSKit compilations, you can launch a BQSKit Runtime
server in either attached or detached mode, and submit a
:class:`~bqskit.compiler.task.CompilationTask` through a connected
:class:`~bqskit.compiler.compiler.Compiler`.

By default, a compiler will automatically start and connect to an attached
runtime server during initialization. In this mode, the compiler
starts up and shuts down the server automatically and transparently to
the end user. See :class:`~bqskit.runtime.attached.AttachedServer` for
more info.

A server running in attached mode is limited to a single machine. If you
would like to parallelize BQSKit across multiple nodes over a network,
you can launch a server independently in detached mode. Here, you first
start :class:`~bqskit.runtime.manager.Manager` processes on all machines.
Then link them all together by running a
:class:`~bqskit.runtime.detached.DetachedServer` process, which Compilers'
can connect to by passing the appropriate address in their constructor.

Upon installing BQSKit, two shell scripts are added to your environment
that can spawn managers and detached servers::

    bqskit-manager

and::

    bqskit-server <address_of_bqskit_managers>

Typically, you start managers first on all nodes in the desired cluster.
Then you can start the server with a comma-separated list of all managers
ip address and optionally ports. Once a server is started and has connected
to the requested managers, no more managers can be added. You can see
the `-h` option of each command or the
:func:`~bqskit.runtime.detached.start_server` and
:func:`~bqskit.runtime.manager.start_manager` entry points for
more information.

For more information on how to manage jobs submitted to a runtime server,
see the :obj:`~bqskit.compiler` documentation.

.. autosummary::
    :toctree: autogen
    :recursive:

    ~bqskit.runtime.attached.AttachedServer
    ~bqskit.runtime.detached.DetachedServer
    ~bqskit.runtime.manager.Manager
    ~bqskit.runtime.detached.start_server
    ~bqskit.runtime.manager.start_manager
    ~bqskit.runtime.worker.start_worker_rank

Parallelize Pass Computation
############################

When developing a BQSKit compiler pass, you can hook into the active
runtime server through the :func:`get_runtime` function. This returns a
:class:`RuntimeHandle`, which you can use to submit, map, wait on, and
cancel tasks in the execution environment.

For more information on how to design a custom pass, see the following
guide: :doc:`guides/custompass.md`.

.. autosummary::
    :toctree: autogen
    :recursive:

    ~bqskit.runtime.get_runtime
    ~bqskit.runtime.RuntimeHandle
    ~bqskit.runtime.future.RuntimeFuture

Standard Logging Interface
##########################

This system uses the standard python logging module to provide a familiar
logging interface. On client processes -- the ones where you create the
Compiler object -- you can configure python loggers like normal before
submitting tasks and have the entire system honor that configuration.
"""
from __future__ import annotations

import faulthandler
import os
from typing import Any
from typing import Callable
from typing import Protocol
from typing import Sequence
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from bqskit.runtime.future import RuntimeFuture


# Enable low-level fault handling: system crashes print a minimal trace.
faulthandler.enable()
os.environ['RUST_BACKTRACE'] = '1'


# Control multi-threading in BLAS libraries.
def set_blas_thread_counts(i: int = 1) -> None:
    """
    Control number of threads used by numpy and others.

    Must be called before any numpy or other BLAS libraries are loaded.
    """
    os.environ['OMP_NUM_THREADS'] = str(i)
    os.environ['OPENBLAS_NUM_THREADS'] = str(i)
    os.environ['MKL_NUM_THREADS'] = str(i)
    os.environ['NUMEXPR_NUM_THREADS'] = str(i)
    os.environ['VECLIB_MAXIMUM_THREADS'] = str(i)


class RuntimeHandle(Protocol):
    """
    A structural type capturing BQSKit Runtime capabilities.

    A compiler pass designer can effortlessly parallelize certain aspects
    of compilation using the :func:`submit` and :func:`map` methods.
    These both will return :class:`~bqskit.runtime.future.RuntimeFuture`
    objects, which can be awaited on using the standard python `await`
    keyword. This can also be used to cancel tasks and wait on specific
    events.

    This should never be constructed directly and only accessed through
    the :func:`get_runtime` function, since this is a structural type
    definition and does not contain any implementation.
    """

    def submit(
        self,
        fn: Callable[..., Any],
        *args: Any,
        task_name: str | None = None,
        log_context: dict[str, str] = {},
        **kwargs: Any,
    ) -> RuntimeFuture:
        """
        Submit a function to the runtime for execution.

        This method schedules the function `fn` to be executed by the
        runtime with the provided arguments `args` and keyword arguments
        `kwargs`. The execution may happen asynchronously.

        Args:
            fn (Callable[..., Any]): The function to be executed.

            *args (Any): Variable length argument list to be passed to
                the function `fn`.

            task_name (str | None): An optional name for the task, which
                can be used for logging or tracking purposes. Defaults to
                None, which will use the function name as the task name.

            log_context (dict[str, str]): A dictionary containing logging
                context information. All log messages produced by the fn
                and any children tasks will contain this context if the
                appropriate level (logging.DEBUG) is set on the logger.
                Defaults to an empty dictionary for no added context.

            **kwargs (Any): Arbitrary keyword arguments to be passed to
                the function `fn`.

        Returns:
            RuntimeFuture: An object representing the future result of
            the function execution. This can be used to retrieve the
            result by `await`ing it.

        Example:
            >>> from bqskit.runtime import get_runtime
            >>>
            >>> def add(x, y):
            ...     return x + y
            >>>
            >>> future = get_runtime().submit(add, 1, 2)
            >>> result = await future
            >>> print(result)
            3

        See Also:
            - :func:`map` for submitting multiple tasks in parallel.
            - :func:`cancel` for cancelling tasks.
            - :class:`~bqskit.runtime.future.RuntimeFuture` for more
                information on how to interact with the future object.
        """
        ...

    def map(
        self,
        fn: Callable[..., Any],
        *args: Any,
        task_name: Sequence[str | None] | str | None = None,
        log_context: Sequence[dict[str, str]] | dict[str, str] = {},
        **kwargs: Any,
    ) -> RuntimeFuture:
        """
        Map a function over a sequence of arguments and execute in parallel.

        This method schedules the function `fn` to be executed by the runtime
        for each set of arguments provided in `args`. Each invocation of `fn`
        will be executed potentially in parallel, depending on the runtime's
        capabilities and current load.

        Args:
            fn (Callable[..., Any]): The function to be executed.

            *args (Any): Variable length argument list to be passed to
                the function `fn`. Each argument is expected to be a
                sequence of arguments to be passed to a separate
                invocation. The sequences should be of equal length.

            task_name (Sequence[str | None] | str | None): An optional
                name for the task group, which can be used for logging
                or tracking purposes. Defaults to None, which will use
                the function name as the task name. If a string is
                provided, it will be used as the prefix for all task
                names. If a sequence of strings is provided, each task
                will be named with the corresponding string in the
                sequence.

            log_context (Sequence[dict[str, str]]) | dict[str, str]): A
                dictionary containing logging context information. All
                log messages produced by the `fn` and any children tasks
                will contain this context if the appropriate level
                (logging.DEBUG) is set on the logger. Defaults to an
                empty dictionary for no added context. Can be a sequence
                of contexts, one for each task, or a single context to be
                used for all tasks.

            **kwargs (Any): Arbitrary keyword arguments to be passed to
                each invocation of the function `fn`.

        Returns:
            RuntimeFuture: An object representing the future result of
            the function executions. This can be used to retrieve the
            results by `await`ing it, which will return a list.

        Example:
            >>> from bqskit.runtime import get_runtime
            >>>
            >>> def add(x, y):
            ...     return x + y
            >>>
            >>> args_list = [(1, 2, 3), (4, 5, 6)]
            >>> future = get_runtime().map(add, *args_list)
            >>> results = await future
            >>> print(results)
            [5, 7, 9]

        See Also:
            - :func:`submit` for submitting a single task.
            - :func:`cancel` for cancelling tasks.
            - :func:`next` for retrieving results incrementally.
            - :class:`~bqskit.runtime.future.RuntimeFuture` for more
                information on how to interact with the future object.
        """
        ...

    def cancel(self, future: RuntimeFuture) -> None:
        """
        Cancel all tasks associated with `future`.

        This is asynchronous and non-pre-emptive. The cancel will return
        immediately, the future it represents will become dead, and at some
        point in the future, all tasks and data associated with the cancelled
        one will be removed from the system.
        """
        ...

    def get_cache(self) -> dict[str, Any]:
        """
        Retrieve worker's local cache.

        In situations where a large or non-easily serializable object is
        needed during the execution of a custom pass, `get_cache` can be used
        to load and store data. Objects placed in cache are transient. Objects
        can be placed in cache to reduce reloading, but should not be expected
        to remain in cache forever.

        For example, say some CustomPassA needs a SomeBigScaryObject. Within
        its run() method, the object is constructed, then placed into cache.
        A later pass, CustomPassB, also needs to use SomeBigScaryObject. It
        checks that the object has been loaded by CustomPassA, then uses it.

        ```
            # Load helper method
            def load_or_retrieve_big_scary_object(file):
                worker_cache = get_runtime().get_cache()
                if 'big_scary_object' not in worker_cache:
                    # Expensive io operation...
                    big_scary_object = load_big_scary_object(file)
                    worker_cache['big_scary_object'] = big_scary_object
                big_scary_object = worker_cache['big_scary_object']
                return big_scary_object

            # In CustomPassA's .run() definition...
            # CustomPassA performs the load.
            big_scary_object = load_or_retrieve_big_scary_object(file)
            big_scary_object.use(...)

            # In CustomPassB's .run() definition...
            # CustomPassB saves time by retrieving big_scary_object from
            # cache after CustomPassA has loaded it.
            big_scary_object = load_or_retrieve_big_scary_object(file)
            unpicklable_object.use(...)
        ```
        """
        ...

    async def next(self, future: RuntimeFuture) -> list[tuple[int, Any]]:
        """
        Wait for and return the next batch of results from a map task.

        Returns:
            (list[tuple[int, Any]]): A list of the results that arrived
                while the task was waiting. Each result is paired with
                the index of its arguments in the original map call.
        """
        ...


def get_runtime() -> RuntimeHandle:
    """
    Return a handle on the active runtime.

    See :class:`RuntimeHandle` for more info.
    """
    from bqskit.runtime.worker import get_worker
    return get_worker()


default_server_port = 7472
default_manager_port = 7473
default_worker_port = 7474
__all__ = ['get_runtime']



"""This module implements numerical functions."""
from __future__ import annotations

from typing import Any

import numpy as np
import numpy.typing as npt
import scipy as sp

from bqskit.qis.pauli import PauliMatrices
from bqskit.qis.pauliz import PauliZMatrices
from bqskit.qis.unitary.unitary import RealVector


def dexpmv(
    M: npt.NDArray[np.complex128], dM: npt.NDArray[np.complex128],
) -> tuple[npt.NDArray[np.complex128], npt.NDArray[np.complex128]]:
    """
    Compute the Matrix exponential F = e^M and its derivative dF.

    User must provide M and its derivative dM. If the argument dM is a
    vector of partials then dF will be the respective partial vector.
    This is done using a Pade Approximation with scaling and squaring.

    Args:
        M (np.ndarray): Matrix to exponentiate.

        dM (np.ndarray): Derivative(s) of M.

    Returns:
        tuple: Tuple containing
            - F (np.ndarray): Exponentiated matrix, i.e. e^M.

            - dF (np.ndarray): Derivative(s) of F.

    References:
        Brančík, Lubomír. "Matlab programs for matrix exponential function
        derivative evaluation." Proc. of Technical Computing Prague 2008
        (2008): 17-24.
    """

    norm = np.linalg.norm(M, np.inf)
    e = np.log2(norm) if norm != 0 else -np.inf
    r = int(max(0, e + 1))
    M = M / (2 ** r)
    dM = dM / (2 ** r)
    X = M
    Y = dM
    c = 0.5
    F = np.identity(M.shape[0]) + c * M
    D = np.identity(M.shape[0]) - c * M
    dF = c * dM
    dD = -c * dM
    q = 6
    p = True
    for k in range(2, q + 1):
        c = c * (q - k + 1) / (k * (2 * q - k + 1))
        Y = dM @ X + M @ Y
        X = M @ X
        cX = c * X
        cY = c * Y
        F = F + cX
        dF = dF + cY
        if p:
            D = D + cX
            dD = dD + cY
        else:
            D = D - cX
            dD = dD - cY
        p = not p
    Dinv = np.linalg.inv(D)
    F = Dinv @ F
    dF = Dinv @ (dF - dD @ F)

    for k in range(1, r + 1):
        dF = dF @ F + F @ dF
        F = F @ F

    return F, dF


def softmax(
    x: npt.NDArray[np.float64],
    beta: int = 20,
) -> npt.NDArray[np.float64]:
    """
    Computes the softmax of vector x.

    Args:
        x (np.ndarray): Input vector to softmax.

        beta (int): Beta coefficient to scale steepness of softmax.

    Returns:
        np.ndarray: Output vector of softmax.
    """

    shiftx = beta * (x - np.max(x))
    exps = np.exp(shiftx)
    return exps / np.sum(exps)


def dot_product(alpha: RealVector, sigma: RealVector) -> npt.NDArray[Any]:
    """
    Computes the standard dot product of `alpha` with `sigma`.

    Args:
        alpha (RealVector): The alpha vector.

        sigma (RealVector): The sigma vector.

    Returns:
        np.ndarray: Sum of element-wise multiplication of `alpha`
        and `sigma`.
    """

    return np.array(np.sum([a * s for a, s in zip(alpha, sigma)], 0))


def unitary_log_no_i(
        U: npt.NDArray[np.complex128],
) -> npt.NDArray[np.complex128]:
    """
    Solves for H in U = e^{iH}

    Args:
        U (np.ndarray): The unitary to decompose.

    Returns:
        np.ndarray: H in e^{iH} = U.

    Note:
        This assumes the input is unitary but does not check. The output
        is undefined on non-unitary inputs.
    """

    T, Z = sp.linalg.schur(U)
    T = np.diag(T)
    D = T / np.abs(T)
    D = np.diag(np.log(D))
    H0 = -1j * (Z @ D @ Z.conj().T)
    return 0.5 * H0 + 0.5 * H0.conj().T


def pauli_expansion(H: npt.NDArray[np.complex128]) -> npt.NDArray[np.float64]:
    """
    Computes a Pauli expansion of the hermitian matrix H.

    Args:
        H (np.ndarray): The hermitian matrix to expand.

    Returns:
        np.ndarray: The coefficients of a Pauli expansion for H,
        i.e., X dot Sigma = H where Sigma is Pauli matrices of
        same size of H.

    Note:
        This assumes the input is hermitian but does not check. The
        output is undefined on non-hermitian inputs.
    """

    # Change basis of H to Pauli Basis (solve for coefficients -> X)
    n = int(np.log2(len(H)))
    paulis = PauliMatrices(n)
    flatten_paulis = [np.reshape(pauli, 4 ** n) for pauli in paulis]
    flatten_H = np.reshape(H, 4 ** n)
    A = np.stack(flatten_paulis, axis=-1)
    X = np.real(np.matmul(np.linalg.inv(A), flatten_H))
    return np.array(X)


def pauliz_expansion(H: npt.NDArray[np.complex128]) -> npt.NDArray[np.float64]:
    """
    Computes a Pauli Z expansion of the diagonal hermitian matrix H.

    Args:
        H (np.ndarray): The (N, N) diagonal hermitian matrix to expand.

    Returns:
        np.ndarray: The coefficients of a Pauli Z expansion for H,
        i.e., x dot Sigma = H where Sigma contains Pauli Z matrices of
        same size of H.

    Note:
        This assumes the input is diagonal but of shape (N, N). No check is
        done for hermicity. The output is undefined on non-hermitian inputs.
    """
    diag_H = np.diag(np.diag(H))
    if not np.allclose(H, diag_H):
        msg = 'H must be a diagonal matrix.'
        raise ValueError(msg)
    # Change basis of H to Pauli Basis (solve for coefficients -> X)
    n = int(np.log2(len(H)))
    paulizs = PauliZMatrices(n)
    flatten_paulizs = [np.diag(pauli) for pauli in paulizs]
    flatten_H = np.diag(H)
    A = np.stack(flatten_paulizs, axis=-1)
    X = np.real(np.matmul(np.linalg.inv(A), flatten_H))
    return np.array(X)


def compute_su_generators(n: int) -> npt.NDArray[np.complex128]:
    """
    Computes the Lie algebra generators for SU(n).

    Args:
        n (int): dimension of SU(n) algebra

    Returns:
        npt.NDArray[np.complex128]: list of the SU(N) generators,
        note that they are Hermitian, but not neccesarily unitary.

    Raises:
        ValueError: if n<=0

    References:
        https://walterpfeifer.ch/liealgebra/LieAlg_wieBuch4.pdf
    """
    # TODO HermitianMatrix objects

    if n <= 0:
        raise ValueError(f'Expected positive integer for n, got: {n}.')

    elif n == 1:
        return np.array([1], dtype=np.complex128)

    elif n == 2:
        return np.array(
            [
                [[0, 1], [1, 0]],
                [[0, -1j], [1j, 0]],
                [[1, 0], [0, -1]],
            ], dtype=np.complex128,
        )

    else:
        previous_generators = compute_su_generators(n - 1)
        generators = [
            np.pad(previous_generators[i], (0, 1))
            for i in range(len(previous_generators))
        ]
        for i in range(n - 1):
            t = np.zeros((n, n), dtype=np.complex128)
            t[i, n - 1] = 1.0
            t[n - 1, i] = 1.0
            generators.append(t)
            t2 = np.zeros((n, n), dtype=np.complex128)
            t2[i, n - 1] = -1j
            t2[n - 1, i] = 1j
            generators.append(t2)

        t3 = np.eye(n)
        t3[n - 1, n - 1] = -n + 1
        t3 *= np.sqrt(2 / (n * (n - 1)))
        generators.append(t3)
        return np.array(generators, dtype=np.complex128)


def canonical_unitary(
    unitary: npt.NDArray[np.complex128],
) -> npt.NDArray[np.complex128]:
    """
    Computes a canonical form for the provided unitary.

    If unitary matrices V, W differ only by a global phase, then
    canonical_unitary(V) == canonical_unitary(W).

    Args:
        unitary (npt.NDArray[np.complex128]): A unitary matrix.

    Returns:
        npt.NDArray[np.complex128]: A unitary matrix.

    References:
        https://arxiv.org/abs/2306.05622
    """
    determinant = np.linalg.det(unitary)
    dimension = len(unitary)
    # Compute special unitary
    global_phase = np.angle(determinant) / dimension
    global_phase = global_phase % (2 * np.pi / dimension)
    global_phase_factor = np.exp(-1j * global_phase)
    special_unitary = global_phase_factor * unitary
    # Standardize speical unitary to account for exp(-i2pi/N) differences
    first_row_mags = np.linalg.norm(special_unitary[0, :], ord=2)
    index = np.argmax(first_row_mags)
    std_phase = np.angle(special_unitary[0, index])
    correction_phase = 0 - std_phase
    std_correction = np.exp(1j * correction_phase)
    return std_correction * special_unitary


def diagonal_distance(unitary: npt.NDArray[np.complex128]) -> float:
    """
    Compute how diagonal a unitary is.

    The diagonal distance measures how closely a unitary can be approx-
    imately inverted by a diagonal unitary. A unitary is approximately
    inverted when the Hilbert-Schmidt distance to the identity is less
    than some threshold.

    A proof of correctness can be found in the appendix of:
    https://arxiv.org/abs/2409.00433

    Args:
        unitary (np.ndarray): The unitary matrix to check.

    Returns:
        float: The Hilbert-Schmidt distance to the nearest diagonal.
    """
    eps = unitary - np.diag(np.diag(unitary))
    eps2 = eps * eps.conj()
    distance = abs(np.sqrt(eps2.sum(-1).max()))
    return distance
